#!/usr/bin/python2

# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Miscellaneous small functions that don't need to be class methods."""

import datetime
import os
# Upstream docs recommend using subprocess32 if possible, but make
# this optional as it's not available in EPEL.
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess
import urllib2

from urlparse import urlparse

def date_check(value, fail_raise=True):
    """Checks whether a value is a date in YYYYMMDD format, and
    returns it as a datetime.datetime instance if so. If raise is set,
    will raise an exception on failure, otherwise returns False. Will
    tolerate ints by forcing them to str.
    """
    try:
        date = datetime.datetime.strptime(str(value), '%Y%m%d')
        return date
    except ValueError:
        if fail_raise:
            err = "{0} is not a valid date in YYYYMMDD format.".format(value)
            raise ValueError(err)
        else:
            return False

def url_exists(url):
    """Checks whether a URL exists, by trying to open it."""
    scheme = urlparse(url).scheme
    if 'http' in scheme:
        try:
            urllib2.urlopen(url)
            return True
        except (ValueError, urllib2.URLError, urllib2.HTTPError):
            return False
    elif 'rsync' in scheme:
        null = open(os.devnull)
        args = ('/usr/bin/rsync', url)
        # Doesn't use check_output as it's not in Python 2.6
        process = subprocess.Popen(args, stdout=null, stderr=null)
        process.communicate()
        retcode = process.poll()
        if retcode:
            return False
        return True

def comma_list(string):
    """Split a comma-separated list of values into a list (used by the
    CLI for passing query parameters etc).
    """
    return string.split(',')

def ci_match(val1, val2, exact, neg):
    """Case-insensitive match, handles either term being false-y. More
    or less expects strings; forces both values to str() to do the
    comparison so it tolerates ints (though it matches them like
    strings). Used by the query stuff.
    """
    if not val1 and not val2:
        ret = True
    elif not val1 or not val2:
        ret = False
    elif exact:
        ret = str(val1).lower() == str(val2).lower()
    else:
        ret = str(val1).lower() in str(val2).lower()
    if neg:
        ret = not ret
    return ret



# -*- coding: utf-8 -*-

__author__ = 'Daniel Greenfeld'
__email__ = 'pydanny@gmail.com'
__version__ = '0.1.5'
__license__ = 'BSD'

import threading


class cached_property(object):
    """ A property that is only computed once per instance and then replaces
        itself with an ordinary attribute. Deleting the attribute resets the
        property.

        Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
        """

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


class threaded_cached_property(cached_property):
    """ A cached_property version for use in environments where multiple
        threads might concurrently try to access the property.
        """
    def __init__(self, func):
        super(threaded_cached_property, self).__init__(func)
        self.lock = threading.RLock()

    def __get__(self, obj, cls):
        with self.lock:
            # Double check if the value was computed before the lock was
            # acquired.
            prop_name = self.func.__name__
            if prop_name in obj.__dict__:
                return obj.__dict__[prop_name]

            # If not, do the calculation and release the lock.
            return super(threaded_cached_property, self).__get__(obj, cls)



#!/usr/bin/python2

# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Defines custom exceptions used by fedfind."""

class NoFilesError(Exception):
    """No files found."""
    pass

class NoImageError(Exception):
    """Files found, but no images."""
    pass



#!/usr/bin/python2

# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""A subclass of koji.ClientSession to handle Koji interaction with a
couple of custom methods.
"""

import datetime
import koji

import fedfind.const
import fedfind.exceptions
import fedfind.helpers

class ClientSession(koji.ClientSession):
    """Add a couple of methods to find nightly image files."""
    def __init__(self, *args, **kwargs):
        super(ClientSession, self).__init__(*args, **kwargs)

    def find_nightly_tasks(self, date=None, release=None, opts=None):
        """This is basically a wrapper around koji's listTasks() which
        will only return tasks that look like nightly image builds. If
        you pass date (as a datetime.date instance) it will find only
        nightlies from that date, by searching a window padded one day
        in either direction and then filtering down by a task request
        parameter which specifies the date. If you pass release (as a
        string, a Fedora release number or 'rawhide') it will find
        only nightlies for that release, by checking another request
        parameter. You can pass opts as a dict of query options, as
        listTasks() expects, and they will be passed through, except
        for 'decode' (which we need to be True to do the filtering),
        'createdAfter' and 'createdBefore' if you pass date (so just
        do one or the other), and 'owner' (which is hardwired to 2745,
        'masher', as the least bad way to avoid treating TC/RC builds
        with a date stamp as nightlies).
        """
        # First, build the query opts dict.
        if not opts:
            opts = dict()
        opts['decode'] = True
        opts['owner'] = 2745
        if date:
            # We'll stretch the window by a day in either direction
            # just in case
            delta = datetime.timedelta(1)
            fromdate = date - delta
            todate = date + delta
            opts['createdAfter'] = fromdate.isoformat()
            opts['createdBefore'] = todate.isoformat()

        # Now find and filter the results. The request parameters -
        # which we get as a list under the key 'request' in the task
        # dict - are not *absolutely* consistent for the different
        # methods, so we have to make some assumptions. We assume that
        # the datestamp and the release are somewhere in the first
        # four parameters.
        tasks = self.listTasks(opts)
        if release:
            tasks = [t for t in tasks
                     if any(str(p) == str(release) for p in t['request'][:3])]
        if date:
            # Filter down to ones for the specific date
            return [t for t in tasks
                    if any(p == date.strftime('%Y%m%d')
                           for p in t['request'][:4])]
        else:
            # Just return any with a version/release that looks like a
            # date (to help filter out any non-nightlies)
            return [t for t in tasks
                    if any(fedfind.helpers.date_check(p, fail_raise=False)
                           for p in t['request'][:4])]

    def find_task_url(self, task):
        """This finds the URL for the image produced by a Koji task,
        if it produced any. If it didn't, it raises an error. It
        assumes a task will only ever produce *one* image; if it
        produces more than one, you'll get the last one Koji gives us.
        """
        # We can't use the file list from task['result'] because it
        # isn't updated when a task's files are aged out, so we can't
        # catch old composes where the files have now been deleted
        # that way.
        files = self.listTaskOutput(task['id'])
        if files:
            imgfile = None
            for fname in files:
                for ext in fedfind.const.IMAGE_EXTS:
                    if fname.endswith(ext):
                        imgfile = fname
            if imgfile:
                url = "{0}/{1}/{2}".format(
                    fedfind.const.KOJI_BASE,
                    koji.pathinfo.taskrelpath(task['id']), imgfile)
                return url
            else:
                # We have some files, but nothing that looks like an
                # image.
                raise fedfind.exceptions.NoImageError
        else:
            # No files - likely a failed or aged-out task.
            raise fedfind.exceptions.NoFilesError



#!/usr/bin/python2

# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Defines the Image class used to represent images."""

import datetime
import fedfind.const
import fedfind.helpers


class Image(object):
    """A representation of a Fedora image. Must be given at least one
    of 'url' or 'path'. 'url' is an absolute URL to a Fedora image.
    'path' is a path relative to the top level of the Fedora mirror
    system - /pub on the master mirror - so if an instance has a 'path'
    attribute,
    https://mirrors.fedoraproject.org/mirrorlist?path=pub/(path)
    should work. If only 'path' is given, 'url' will be set to the
    URL for the image using the https://download.fedoraproject.org/
    redirector. If only 'url' is given, we will figure out 'path' only
    if the URL uses dl.fedoraproject or download.fedoraproject.

    url is used as the string representation, so the string for an
    Image instance should usually be a valid link to download it.

    If both 'url' and 'path' are passed, self.path and self.url will
    both be set, and on your own head be it if for some reason they're
    entirely different. I can't think of a good reason to do this.

    If given only 'url' and/or 'path', will attempt to figure out
    various attributes of the image. Most of the attributes can
    also be overridden at instantiation time, so if you know any of
    them for sure, you can just set them and avoid the guesswork.

    Really should be given a valid HTTPS URL or Fedora mirror system
    path. You may be able to get useful results in some sense by
    passing something else as url and/or path (e.g. a filesystem path)
    but it's not really expected. Can be used fairly independently of
    the rest of fedfind if you just want to use it to figure out some
    attributes of an arbitrary Fedora image.
    """
    def __init__(self, url='', path='', arch='', imagetype='', flavor='',
                 subflavor='', loadout='', date='', release='', milestone='',
                 compose=''):
        if not url and not path:
            raise TypeError("fedfind.Image() needs url or path.")
        self.url = url
        if not url:
            self.url = '{0}/{1}'.format(fedfind.const.HTTPS, path)
        self.path = path
        if not path:
            for pfx in fedfind.const.KNOWN_PREFIXES:
                if url.startswith(pfx):
                    self.path = url.replace(pfx, '', 1)
        self.filename = self.url.split('/')[-1]
        self.arch = arch
        self.imagetype = imagetype
        self.flavor = flavor
        self.subflavor = subflavor
        self.loadout = loadout
        self.payload = ''
        self.date = date
        self.release = release
        self.milestone = milestone
        self.compose = compose

        # attempt to figure out unspecified attributes.
        checkurl = self.url.lower()
        checkfile = self.filename.lower()
        if not arch:
            for arch in fedfind.const.ARCHES:
                if arch in checkurl:
                    self.arch = arch

        if not imagetype:
            if checkfile == 'boot.iso':
                self.imagetype = 'boot'
            if any(checkfile.endswith(ext) for ext in fedfind.const.ISO_EXTS):
                for isotype in fedfind.const.ISO_TYPES:
                    if isotype in checkurl:
                        self.imagetype = isotype
            elif any(checkfile.endswith(ext)
                     for ext in fedfind.const.DISK_EXTS):
                self.imagetype = 'disk'
            elif any(checkfile.endswith(ext)
                     for ext in fedfind.const.DOCKER_EXTS):
                self.imagetype = 'docker'

        if not flavor and not loadout:
            for (flavor, attrs) in fedfind.const.FLAVORS.items():
                for name in attrs['names']:
                    if name in checkurl:
                        self.flavor = flavor
                        if 'subs' in attrs:
                            for sub in attrs['subs']:
                                if '{0}-{1}'.format(name, sub) in checkurl:
                                    self.subflavor = sub
            for (loadout, attrs) in fedfind.const.LOADOUTS.items():
                for name in attrs['names']:
                    if name in checkurl:
                        self.loadout = name

        if self.flavor:
            self.payload = self.flavor
        elif self.loadout:
            self.payload = self.loadout

        elems = checkfile.split('.')[0].split('-')
        for elem in elems:
            # We want to capture the *first* matching element, to get
            # '21' not '5' for Fedora-Live-Workstation-x86_64-21-5.iso
            if (elem.isdigit() and
                    int(elem) in range(0, 100) and not
                    self.release):
                self.release = elem
            try:
                datetime.datetime.strptime(elem, '%Y%m%d')
                if not self.date:
                    self.date = elem
            except:
                pass

    def __str__(self):
        return self.url

    def __repr__(self):
        return ("{0}(url='{1}', path='{2}', arch='{3}', imagetype='{4}', "
                "flavor='{5}', subflavor='{6}', loadout='{7}', date='{8}', "
                "release='{9}')").format(
                    self.__class__, self.url, self.path, self.arch,
                    self.imagetype, self.flavor, self.subflavor, self.loadout,
                    self.date, self.release)

    @property
    def exists(self):
        """Does the image in question actually exist? Try and access
        it and find out. If self.path is set it **MUST** be the path
        relative to the top level of the mirror system. If path is not
        set, url must be a complete https or rsync URL."""
        if self.path:
            rsync = '{0}/{1}'.format(fedfind.const.RSYNC, self.path)
            return fedfind.helpers.url_exists(rsync)
        return fedfind.helpers.url_exists(self.url)



#!/usr/bin/python2

# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Defines the various release classes used as the main entry
points.
"""

import abc
import koji
import os
# Upstream docs recommend using subprocess32 if possible, but make
# this optional as it's not available in EPEL.
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess

from operator import attrgetter

import fedfind.const
import fedfind.exceptions
import fedfind.helpers
import fedfind.image
import fedfind.kojiclient

from fedfind.cached_property import cached_property

def get_release(release='', date=None, milestone='', compose=''):
    """Function to return an appropriate Release subclass for the
    version information. Figures out if you want a nightly, compose,
    milestone, or stable release and gives it to you. Catches some
    nonsense choices. Tries to figure out the appropriate subclass
    (current, archive, core...) for stable releases. Won't do any
    clever guessing of 'current release' or date or anything like
    that. Put that in callers if you want it.
    """
    # Handle date (nightly) cases first.
    if date and (str(release).lower() == 'rawhide' or not release):
        return RawhideNightly(date)
    if date and release:
        return BranchedNightly(date, release)

    # We know we have no date; we can error out if 'release' is wrong.
    if not release:
        raise TypeError("get_release() needs at least release or date.")
    if release == 'rawhide':
        raise TypeError("get_release(): release 'rawhide' needs a date.")

    # Now we've got a release that's not 'rawhide' but no date. First,
    # handle stable releases. milestone 'Final' and no compose ==
    # final release.
    if not compose and (not milestone or milestone == 'Final'):
        if int(release) < 7:
            return CoreRelease(release)
        if int(release) < 20:
            return ArchiveRelease(release)
        # All the ways we can try to handle the stable vs. archived
        # cutoff are basically hideous. This is the one I'm picking.
        rel = ArchiveRelease(release)
        if not rel.exists:
            rel = CurrentRelease(release)
        return rel

    # By ze Process of Elimination, we've now got a release and at
    # least one of milestone and compose.
    if not milestone:
        raise TypeError("get_release(): compose without milestone is "
                        "nonsense. Stoppit.")
    if not compose:
        return Milestone(release, milestone)
    return Compose(release, milestone, compose)

class Query(object):
    """Small class for image queries. attr is the name of the object
    attribute against which this query should be run. Terms is an
    iterable of search terms. exact and neg are query parameters.
    If exact is True the match will succeed only on an exact match of
    any query term vs. the object attribute, otherwise the match will
    succeed if any query term is a substring of the object attribute.
    If neg is True, the result is reversed (turns IS into IS NOT and
    IN into NOT IN). Two false-y values are considered a match - if
    the query term is False or None or '' and the object attribute is
    False or None or '', the match succeeds. The object not having the
    attribute at all is treated identically to have it set to a falsey
    value."""
    def __init__(self, attr, terms, exact=True, neg=False):
        self.attr = attr
        self.getter = attrgetter(attr)
        self.terms = terms
        self.exact = exact
        self.neg = neg

    def match(self, obj):
        """Perform the match (for details see class docstring)."""
        try:
            objval = self.getter(obj)
        except AttributeError:
            objval = None
        return any(fedfind.helpers.ci_match(t, objval, self.exact, self.neg)
                   for t in self.terms)

class Release(object):
    """Abstract class for releases, providing shared attributes and
    methods.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, date=None, release='', milestone='', compose=''):
        self.release = release
        self.date = date
        self.milestone = milestone
        self.compose = compose

    @abc.abstractproperty
    def all_images(self):
        """All images for this release."""
        pass

    @abc.abstractproperty
    def exists(self):
        """Whether the release is considered to 'exist'."""
        pass

    def find_images(self, queries=None, orq=False):
        """This is a simple 'fallback' query method which simply finds
        all images for the release and runs the query function. The
        idea is that subclasses can have more sophisticated
        find_images() methods which try to reduce the set of images
        queried when possible to improve performance. find_images()
        should always ultimately call _query_images() on the set of
        images it decides on, with the iterable of queries it was
        passed, to handle the query stuff. See the _query_images()
        docstring for details on queries and orq.
        """
        if not queries:
            return self.all_images
        return self._query_images(queries, orq=orq)

    def _query_images(self, queries, imgs='all', orq=False):
        """This is a very generic method for searching a given set of
        images. queries is an iterable of Query objects (or I guess
        an iterable whose members can be any objects with a 'match'
        method that you want to use to filter the image list). Refer
        to the Query class docstrings for more details on how to
        create appropriate queries.

        imgs is an iterable of fedfind.Image objects to query. If left
        at its default value, all images (the all_images property)
        will be queried.

        Multi-attribute searches are AND by default - only images for
        which every query.match() call passes will be returned. You
        can pass orq=True to change this behavior to OR, in which case
        images for which *any* query.match() call passes will be
        returned.
        """
        if orq:
            orset = set()
        if imgs == 'all':
            imgs = self.all_images
        for query in queries:
            match = [i for i in imgs if query.match(i)]
            if orq:
                orset.update(match)
            else:
                imgs = match
        if orq:
            imgs = list(orset)
        return imgs

    def image_from_url_or_path(self, url='', path=''):
        """This gets an image instance for a given URL or mirror
        path, setting any properties that we know from the release's
        properties.
        """
        if not url and not path:
            raise TypeError("image_from_url_or_path() needs url or path.")
        return fedfind.image.Image(
            url=url, path=path, release=self.release, date=self.date,
            milestone=self.milestone, compose=self.compose)

## NIGHTLY-TYPE RELEASE CLASSES ##

class Nightly(Release):
    """Parent class for nightly releases: Rawhide and Branched
    nightlies are very similar and share lots of logic, only a few
    details differ in the child classes. Should not be instantiated
    directly.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, date, release):
        super(Nightly, self).__init__(date=date, release=release)

    @abc.abstractproperty
    def _mashbase(self):
        """The top-level URL for this release in mash."""
        pass

    @cached_property
    def all_boot_images(self):
        """All boot.iso-type images. We can get these without hitting
        Koji.
        """
        imgs = list()
        for arch in fedfind.const.NIGHTLY_BOOT_ARCHES:
            img = self.image_from_url_or_path(url=self._get_boot_url(arch))
            if img.exists:
                imgs.append(img)
        return imgs

    @cached_property
    def all_koji_images(self):
        """All images we have to get from Koji."""
        return self.get_koji_images()

    @cached_property
    def all_images(self):
        """All images for this release."""
        return self.all_boot_images + self.all_koji_images

    @cached_property
    def exists(self):
        return fedfind.helpers.url_exists(self._mashbase)

    def _get_boot_url(self, arch):
        """Get the expected URL for the boot.iso of a given arch."""
        tmpl = '{0}/{1}/os/images/boot.iso'
        return tmpl.format(self._mashbase, arch)

    def _check_koji_urls(self, opts=None):
        """Query Koji for nightly tasks, and return a list of image
        URLs. By default finds all live images. opts is an opts dict
        for the koji listTasks() method; the purpose of passing one
        from here is to specify the method we want to find images
        for. Used by get_koji_images()."""
        client = fedfind.kojiclient.ClientSession(
            'http://koji.fedoraproject.org/kojihub', {})
        if not opts:
            opts = dict()
        urls = list()
        opts['states'] = [koji.TASK_STATES['CLOSED']]
        if 'method' not in opts:
            opts['method'] = 'createLiveCD'
        tasks = client.find_nightly_tasks(
            date=self.date, release=self.release, opts=opts)
        for task in tasks:
            try:
                urls.append(client.find_task_url(task))
            except (fedfind.exceptions.NoImageError,
                    fedfind.exceptions.NoFilesError):
                pass
        return urls

    def get_koji_images(
            self, methods=('createLiveCD', 'createAppliance', 'createImage'),
            arches=None):
        """Find all or some of our Koji nightly tasks, and return a
        list of fedfind.Image instances, one per task that produced
        images. Methods is an iterable of Koji task methods (each
        method requires another query, so we want to be able to limit
        this to as few methods as possible when we can).
        """
        imgs = list()
        for method in methods:
            opts = dict()
            opts['method'] = method
            if arches:
                opts['arch'] = list(arches)
            imgs.extend([self.image_from_url_or_path(url=url)
                         for url in self._check_koji_urls(opts)])
        return imgs

    def find_images(self, queries=None, orq=False):
        """This implementation tries to speed up certain queries by
        avoiding some remote queries where possible. If we are only
        asked for boot images, it doesn't hit Koji. If we *aren't*
        asked for boot images, it doesn't hit the mirrors. If we are
        only asked for certain types of image, it restricts the list
        of methods we query Koji for (each method requires an extra
        query).
        """
        if not queries:
            return self.all_images
        arches = None
        bootimgs = kojiimgs = list()
        needboot = needlive = needdisk = True
        for query in queries:
            # We can only narrow down the external queries if the fedfind query
            # is positive and exact.
            if query.attr == 'imagetype' and not query.neg and query.exact:
                if 'boot' not in query.terms:
                    needboot = False
                if 'live' not in query.terms:
                    needlive = False
                if 'disk' not in query.terms:
                    needdisk = False
                arches = query.terms

        if needboot:
            bootimgs = self.all_boot_images

        methods = list()
        if needlive and needdisk and not arches:
            kojiimgs = self.all_koji_images

        if needlive:
            methods.append('createLiveCD')
        if needdisk:
            methods.extend(['createAppliance', 'createImage'])
        if methods and not kojiimgs:
            kojiimgs = self.get_koji_images(methods=methods, arches=arches)

        imgs = bootimgs+kojiimgs
        return self._query_images(queries, imgs, orq=orq)


class RawhideNightly(Nightly):
    """A Rawhide nightly 'release', consisting of the nightly composes
    for the Rawhide branch for the specified date.
    """
    def __init__(self, date):
        release = 'rawhide'
        super(RawhideNightly, self).__init__(date, release)

    @property
    def _mashbase(self):
        return '{0}/rawhide-{1}/rawhide'.format(
            fedfind.const.MASH_BASE, self.date.strftime('%Y%m%d'))


class BranchedNightly(Nightly):
    """A Branched nightly 'release', consisting of the nightly composes
    for the Branched branch for the specified date. Given how the
    mash directory is laid out, you have to specify a release; it's
    not feasible to just find 'the' Branched 'release' for a given
    date. Callers that want to be helpful could make a sensible guess
    like 'current release plus one', though - that's usually the only
    one you'll be interested in and able to find any images for.
    """
    def __init__(self, date, release):
        release = str(release)
        super(BranchedNightly, self).__init__(date, release)

    @property
    def _mashbase(self):
        return '{0}/branched-{1}/{2}'.format(
            fedfind.const.MASH_BASE, self.date.strftime('%Y%m%d'),
            self.release)

## END NIGHTLY-TYPE RELEASE CLASSES ##

class MirrorRelease(Release):
    """A parent class for releases for which all files are on the
    mirrors (no koji or mash stuff)."""
    __metaclass__ = abc.ABCMeta

    def __init__(self, release, milestone='', compose=''):
        super(MirrorRelease, self).__init__(
            release=release, milestone=milestone, compose=compose)

    @abc.abstractproperty
    def _rsyncpath(self):
        """Top level path for this release, relative to the fedora-
        buffet bundle in rsync (/pub on the master mirror).
        """
        pass

    @cached_property
    def all_images(self):
        """All images for this release."""
        return self.get_mirror_images()

    @cached_property
    def exists(self):
        """Release 'exists' if its top-level mirror path is there."""
        url = '{0}/{1}'.format(fedfind.const.RSYNC, self._rsyncpath)
        return fedfind.helpers.url_exists(url)

    def get_mirror_images(self):
        """Find images in the main mirror tree by parsing rsync output.
        This is about the simplest / most reliable method I could
        figure out for scraping the mirror tree. _rsyncpath is the path
        prefix to restrict the search to.
        """
        # This excludes files but includes directories (we have to
        # include directories to be able to include their contents)
        args = ['/usr/bin/rsync', '{0}/{1}/'.format(fedfind.const.RSYNC,
                                                    self._rsyncpath),
                '--recursive', '--exclude="*"', '--include="*/"']
        # This includes files that end with one of our image
        # extensions
        args.extend(['--include="*{0}"'.format(ext)
                     for ext in fedfind.const.IMAGE_EXTS])
        null = open(os.devnull)
        # Doesn't use check_output as it's not in Python 2.6
        process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=null)
        out = process.communicate()[0]
        retcode = process.poll()
        if retcode:
            return []
        # This finds the output lines that are image files (ones
        # ending with our image extensions) and produces a URL by
        # combining the base alt URL with the file path, which is the
        # fifth field of the rsync output
        paths = ['{0}/{1}'.format(self._rsyncpath, o.split()[4])
                 for o in out.splitlines()
                 for e in fedfind.const.IMAGE_EXTS if o.endswith(e)]
        return [self.image_from_url_or_path(path=path) for path in paths]


class Compose(MirrorRelease):
    """A TC/RC compose, stored in the staging tree."""
    def __init__(self, release, milestone, compose):
        super(Compose, self).__init__(release=release, milestone=milestone,
                                      compose=compose)

    @property
    def _rsyncpath(self):
        if self.milestone == 'Final':
            return 'alt/stage/{0}_{1}'.format(self.release, self.compose)
        else:
            return 'alt/stage/{0}_{1}_{2}'.format(
                self.release, self.milestone, self.compose)

class Milestone(MirrorRelease):
    """A milestone release - Alpha or Beta."""
    def __init__(self, release, milestone):
        super(Milestone, self).__init__(release=release, milestone=milestone)

    @property
    def _rsyncpath(self):
        return 'fedora/linux/releases/test/{0}-{1}'.format(
            self.release, self.milestone)

## STABLE RELEASE CLASSES ##

class CurrentRelease(MirrorRelease):
    """A release that public mirrors are expected to carry (not an
    archived one).
    """
    def __init__(self, release):
        super(CurrentRelease, self).__init__(release=release)

    @property
    def _rsyncpath(self):
        return 'fedora/linux/releases/{0}'.format(self.release)

class ArchiveRelease(MirrorRelease):
    """An archived Fedora (post-Fedora Core) release."""
    def __init__(self, release):
        super(ArchiveRelease, self).__init__(release=release)

    @property
    def _rsyncpath(self):
        return 'archive/fedora/linux/releases/{0}'.format(self.release)


class CoreRelease(MirrorRelease):
    """A Fedora Core release."""
    def __init__(self, release):
        super(CoreRelease, self).__init__(release=release)

    @property
    def _rsyncpath(self):
        return 'archive/fedora/linux/core/{0}'.format(self.release)

## END STABLE RELEASE CLASSES ##



#!/usr/bin/python2

# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""various shared constants."""

KOJI_BASE = 'https://kojipkgs.fedoraproject.org/work'
DISK_EXTS = ('.raw.xz', '.qcow2')
DOCKER_EXTS = ('.tar.gz',)
ISO_EXTS = ('.iso',)
IMAGE_EXTS = DISK_EXTS + DOCKER_EXTS + ISO_EXTS
NIGHTLY_BOOT_ARCHES = ('x86_64', 'i386', 'armhfp')
ARCHES = ('x86_64', 'i386', 'i686', 'armhfp', 'ppc', 'source')
RSYNC = 'rsync://dl.fedoraproject.org/fedora-buffet'
HTTPS = 'https://download.fedoraproject.org/pub'
KNOWN_PREFIXES = (RSYNC, HTTPS, 'http://download.fedoraproject.org/pub',
                  'https://dl.fedoraproject.org/pub',
                  'http://dl.fedoraproject.org/pub')
MASH_BASE = 'https://kojipkgs.fedoraproject.org/mash'

FLAVORS = dict(cloud = dict(names=('cloud',), subs=('base', 'atomic')),
               workstation = dict(names=('workstation',)),
               server = dict(names=('server',)))
LOADOUTS = dict(desktop = dict(names=('desktop',)),
                kde = dict(names=('kde',)),
                xfce = dict(names=('xfce',)),
                lxde = dict(names=('lxde',)),
                mate = dict(names=('mate_compiz', 'mate-compiz', 'mate')),
                soas = dict(names=('soas',)),
                design_suite = dict(names=('design_suite', 'design-suite')),
                electronic_lab = dict(names=('electronic_lab', 'electronic-lab')),
                games = dict(names=('games',)),
                jam_kde = dict(names=('jam_kde',)),
                robotics = dict(names=('robotics',)),
                scientific_kde = dict(names=('scientific_kde', 'scientific-kde')),
                security = dict(names=('security',)),
                minimal = dict(names=('minimal',)))
ISO_TYPES = ('dvd', 'live', 'netinst', 'disc1', 'disc2', 'disc3', 'disc4', 'disc5', 'disc6')



import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below. Stolen from
# https://pythonhosted.org/an_example_pypi_project/setuptools.html
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "fedfind",
    version = "1.0.3",
    entry_points = {'console_scripts': ['fedfind = fedfind.cli:main'],},
    author = "Adam Williamson",
    author_email = "awilliam@redhat.com",
    description = "Fedora Finder finds Fedora - images, more to come?",
    license = "GPLv3+",
    keywords = "fedora release image media iso",
    url = "https://www.happyassassin.net/fedfind/",
    packages = ["fedfind"],
    install_requires = ['subprocess32', 'setuptools'],
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 or later "
        "(GPLv3+)",
    ],
)



#!/usr/bin/python2

# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""CLI module for fedfind."""

import argparse
import sys

import fedfind.helpers
import fedfind.release

class Cli(object):
    """Bits for the fedfind CLI tool."""
    def __init__(self):
        # Define 'valid' TC/RC names for arg parsing
        self._composes = set(
            '{0}{1}'.format(a, b) for a in ('TC', 'RC') for b in range(1, 21))
        # Parse command-line args
        self.args = self.parse_args()

    def images(self):
        """images sub-command method. Gets an appropriate Release
        instance and runs an appropriate query, based on the provided
        args.
        """
        try:
            rel = fedfind.release.get_release(
                release=self.args.release, date=self.args.date,
                milestone=self.args.milestone, compose=self.args.compose)
        except TypeError:
            sys.exit("Invalid arguments. You must pass --date or --release, "
                     "and you cannot pass --compose without --milestone.")
        queries = list()
        if self.args.arch:
            if 'i386' in self.args.arch and not 'i686' in self.args.arch:
                self.args.arch.append('i686')
            elif 'i686' in self.args.arch and not 'i386' in self.args.arch:
                self.args.arch.append('i386')
            queries.append(fedfind.release.Query('arch', self.args.arch))
        if self.args.type:
            queries.append(fedfind.release.Query('imagetype', self.args.type))
        if self.args.payload:
            queries.append(fedfind.release.Query('payload', self.args.payload))
        if self.args.search:
            queries.append(
                fedfind.release.Query('url', self.args.search, exact=False))
        imgs = rel.find_images(queries)
        for img in imgs:
            print(img)

    def parse_args(self):
        """Parse arguments with argparse."""
        parser = argparse.ArgumentParser(description=(
            "Tool for finding Fedora stuff. Currently finds images, using the "
            "images sub-command. See the image help for more details."))
        subparsers = parser.add_subparsers()
        parser_images = subparsers.add_parser(
            'images', description="Find Fedora images. You must identify a "
            "release, using at least --release or --date. If you specify only "
            "--release, it must be a Fedora release number, and fedfind will "
            "find images for that stable release. If you specify only --date, "
            "it must be a date in YYYYMMDD format, and fedfind will find "
            "Rawhide nightly images for that date. If you specify --release NN"
            " --date YYYYMMDD, fedfind will find Branched nightly images - you"
            " have to provide the release to search Branched. If you specify "
            "--release and --milestone, fedfind will find images for a pre-"
            "release - e.g. --release 22 --milestone Alpha. If you specify "
            "--release, --milestone and --compose, fedfind will find images "
            "for a TC/RC compose - e.g. --release 22 --milestone Alpha "
            "--compose TC1. The other parameters can be used to filter the "
            "results in various ways. For all of the query parameters, you can"
            " specify a single value or several separated by commas. If you "
            "specify several values, fedfind will find images that match *any*"
            " value. If you pass multiple query parameters, fedfind will only "
            "find images that pass the check for all of the parameters. "
            "Matching is exact (but not case-sensitive) for all parameters "
            "except --search, which will match any image where at least one "
            "of the search terms occurs somewhere in the image URL.")
        parser_images.add_argument(
            '-r', '--release', help="The Fedora release to search", type=int,
            required=False, choices=range(1, 100), metavar="1-99")
        parser_images.add_argument(
            '-d', '--date', help="The date of a nightly compose to search",
            type=fedfind.helpers.date_check, required=False,
            metavar="YYYYMMDD")
        parser_images.add_argument(
            '-m', '--milestone', help="A milestone to search (e.g. Alpha or "
            "Beta)", choices=['Alpha', 'Beta', 'Final'], required=False)
        parser_images.add_argument(
            '-c', '--compose', help="A compose to search (e.g. TC1 or RC3)",
            choices=self._composes, metavar="{T,R}C1-19", required=False)
        parser_images.add_argument(
            '-a', '--arch', help="Architecture(s) to search for",
            required=False, type=fedfind.helpers.comma_list,
            metavar="armhfp,x86_64...")
        parser_images.add_argument(
            '-t', '--type', help="Image type(s) to search for", required=False,
            metavar="boot,netinst,live...", type=fedfind.helpers.comma_list)
        parser_images.add_argument(
            '-p', '--payload', help="Image payload to search for",
            metavar="workstation,lxde...", required=False,
            type=fedfind.helpers.comma_list)
        parser_images.add_argument(
            '-s', '--search', help="String(s) to search for anywhere in image "
            "URL", required=False, type=fedfind.helpers.comma_list,
            metavar="TERM1,TERM2")
        parser_images.set_defaults(func=self.images)
        return parser.parse_args()

    def run(self):
        """Read in arguments and run sub-command function."""
        self.args.func()

def main():
    """Main loop."""
    try:
        cli = Cli()
        cli.run()
    except KeyboardInterrupt:
        print "Interrupted, exiting..."
        sys.exit(1)

if __name__ == '__main__':
    main()
